import redis
import logging
import json

from .preprocess import preprocess_file, preprocess_browse
from .exceptions import ExistsAtUploadError

logger = logging.getLogger(__name__)


def run_daemon(config, host, port, listen_queue, listen_md_queue, write_queue, progress_set, failure_set, success_set):
    """ Run the preprocessing daemon, listening on a redis queue
        for files to be preprocessed. After preprocessing the filename
        of the preprocessed files will be pushed to the output queue.
    """
    # initialize the queue client
    client = redis.Redis(
        host=host, port=port, charset="utf-8", decode_responses=True
    )
    logger.debug("waiting for redis queue '%s'..." % listen_queue)
    while True:
        # fetch an item from the queue to be preprocessed
        queue, value = client.brpop([listen_queue, listen_md_queue])
        client.sadd(progress_set, value)
        file_paths = []
        # start the preprocessing on that file
        if queue == listen_queue:
            try:
                filename, file_path = preprocess_file(config, value)
                file_paths.append(file_path)
                client.sadd(success_set, value)
                client.srem(progress_set, value)
            except ExistsAtUploadError as e:
                # do not add file already present to failure set
                client.srem(progress_set, value)
                logger.exception(e)
                continue
            except Exception as e:
                client.sadd(failure_set, value)
                client.srem(progress_set, value)
                logger.exception(e)
                continue
        elif queue == listen_md_queue:
            browse_report_data = json.loads(value)
            browse_type = browse_report_data['browse_type']
            for browse in browse_report_data['browses']:
                try:
                    filename, file_path = preprocess_browse(config, browse_type, browse_report_data, browse)
                    file_paths.append(file_path)
                    client.sadd(success_set, browse['filename'])
                except ExistsAtUploadError as e:
                    # do not add file already present to failure set
                    logger.exception(e)
                    continue
                except Exception as e:
                    client.sadd(failure_set, browse['filename'])
                    logger.exception(e)
                    continue
        client.srem(progress_set, value)  # for browse_report mode
        # TODO: convert to string, list, ....
        for item in file_paths:
            client.lpush(write_queue, item)
